/****************************************************************************
**
*A  costab.h                    GAP source                   Martin Schoenert
*A                                                           & Volkmar Felsch
**
*A  @(#)$Id: costab.h,v 3.1 1992/11/16 18:55:27 martin Rel $
**
*Y  Copyright 1990-1992,  Lehrstuhl D fuer Mathematik,  RWTH Aachen,  Germany
**
**  This file defines the functions for computing with coset tables.
**
*H  $Log: costab.h,v $
*H  Revision 3.1  1992/11/16  18:55:27  martin
*H  initial revision under RCS
*H
*/


/****************************************************************************
**
*F  CompressDeductionList( )  . . .  removes unused items from deduction list
**
**  'CompressDeductionList'  tries to find and delete  deduction list entries
**  which are not used any more.
**
**  'dedgen',  'dedcos',  'dedfst',  'dedlst',  'dedSize'  and 'ptTable'  are
**  assumed to be known as static variables.
*/
extern  void            CompressDeductionList P(( void ));


/****************************************************************************
**
*F  FunApplyRel( <hdCall> ) . . . . . . .  apply a relator to a coset in a TC
**
**  'FunApplyRel' implements the internal function 'ApplyRel'.
**
**  'ApplyRel( <app>, <rel> )'
**
**  'ApplyRel'  applies the relator  <rel>  to the  application  list  <app>.
**  ...more about ApplyRel...
*/
extern  TypHandle       FunApplyRel P(( TypHandle hdCall ));


/****************************************************************************
**
*F  HandleCoinc(<cos1>,<cos2>) . . . . . . . . .  handle coincidences in a TC
**
**  'HandleCoinc'  is a subroutine of  'FunMakeConsequences'  and handles the
**  coincidence  cos2 = cos1.
*/
extern  void            HandleCoinc P(( unsigned long, unsigned long ));


/*****************************************************************************
**
*F  FunMakeConsequences(<hdCall>) . . find consequences of a coset definition
*/
extern  TypHandle       FunMakeConsequences P(( TypHandle hdCall ));


/****************************************************************************
**
*F  FunStandardizeTable(<hdCall>)  . . . . . . . .  standardize a coset table
*/
extern  TypHandle       FunStandardizeTable P(( TypHandle hdCall ));


/****************************************************************************
**
*F  FunApplyRel2( <hdCall> ) . . . .  apply a relator to a coset rep in a CRT
**
**  'FunApplyRel2' implements the internal function 'ApplyRel2'.
**
**  'ApplyRel2( <app>, <rel>, <nums> )'
**
**  'ApplyRel2'  applies  the relator  <rel>  to a  coset representative  and
**  returns the corresponding factors in "word"
**  ...more about ApplyRel2...
*/
extern  TypHandle       FunApplyRel2 P(( TypHandle hdCall ));


/****************************************************************************
**
*F  FunCopyRel( <hdCall> ) . . . . . . . . . . . . . . . .  copy of a relator
**
**  'FunCopyRel'  returns a copy of the given  RRS relator  such that the bag
**  of the copy does not exceed the minimal required size.
*/
extern  TypHandle       FunCopyRel P(( TypHandle hdCall ));


/****************************************************************************
**
*F  FunMakeCanonical( <hdCall> ) . . . . . . . . . . make a relator canonical
**
**  'FunMakeCanonical'  is a subroutine of the  Reduced Reidemeister-Schreier
**  routines.  It replaces the given relator by its canonical representative.
**  It does not return anything.
*/
extern  TypHandle       FunMakeCanonical P(( TypHandle hdCall ));


/****************************************************************************
**
*F  FunTreeEntry( <hdCall> ) . . . .  returns a tree entry for the given word
**
**  'FunTreeEntry'  determines a  tree entry  which represents the given word
**  in the current generators,  if it finds any,  or it defines a  new proper
**  tree entry, and then returns it.
*/
extern  TypHandle       FunTreeEntry P(( TypHandle hdCall ));


/****************************************************************************
**
*F  TreeEntryC( )  . . . . . . . . . . .  returns a tree entry for a rep word
**
**  'TreeEntryC'  determines a tree entry  which represents the word given in
**  'wordList', if it finds any, or it defines a  new proper tree entry,  and
**  then returns it.
**
**  Warning:  It is assumed,  but not checked,  that the given word is freely
**  reduced and that it does not contain zeros.
**
**  'wordList'  is  assumed to be known as static variable.
**
*/
extern  long            TreeEntryC P(( void ));


/****************************************************************************
**
*F  AddCosetFactor( <factor> ) .  add a factor to a coset representative word
**
**  'AddCosetFactor' adds a factor to a coset representative word and extends
**  the tree appropriately, if necessary.
**
**  'wordList'  and  'wordSize'  are assumed to be known as static variables.
**  Warning: 'factor' is not checked for being zero.
*/
extern  void            AddCosetFactor P(( long ));


/****************************************************************************
**
*F  HandleCoinc2(<cos1>,<cos2>,<factor>)  . . .  handle coincidences in a MTC
**
**  'HandleCoinc2'  is a subroutine of 'FunMakeConsequences2' and handles the
**  coincidence  cos2 = factor * cos1.
*/
extern  void            HandleCoinc2 P(( long, long, long ));


/****************************************************************************
**
*F  FunMakeConsequences2(<hdCall>) .  find consequences of a coset definition
*/
extern  TypHandle       FunMakeConsequences2 P(( TypHandle hdCall ));


/****************************************************************************
**
*F  FunStandardizeTable2(<hdCall>) . . . . . . . . . . . .  standardize a CRT
*/
extern  TypHandle       FunStandardizeTable2 P(( TypHandle hdCall ));


/****************************************************************************
**
*F  InitCostab()  . . . . . . . . . . . .  initialize the coset table package
**
**  'InitCostab' initializes the coset table package.
*/
extern  void            InitCosTab P(( void ));



